/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * I don't know who the original author of this program is, but the
 * module-stuff is done by me. 
 *		Jens Ch. Restemeier <jchrr@hrz.uni-bielefeld.de>
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <dirent.h>
#ifdef __linux__
#include <linux/vt.h>
#endif
#include <string.h>
#include <ggi/libggi.h>
#include "ggi_screensaver.h"

/* for dynamic linking */
#include <dlfcn.h>

/*
 * Module handling:
 */

ggi_saver_list *rs_list;
int rs_list_num, rs_list_size;
char **m_list;
int m_list_num, m_list_size;

volatile int switchaway;

void install_saver(saver_func *func, char *name)
{
	if (rs_list_size==0) {
		rs_list_size=10;
		rs_list=malloc(sizeof(ggi_saver_list)*rs_list_size);
	}
	if (rs_list_num==rs_list_size) {
		rs_list_size+=10;
		rs_list=realloc(&rs_list, sizeof(ggi_saver_list)*rs_list_size);
	}
	rs_list[rs_list_num].name=name;
	rs_list[rs_list_num].func=func;
	rs_list_num++;
}

void read_module(char *s)
{
	void *h;
	char *error;
	char tmp_path[256];
	if (s[0]=='~') {
		strcpy(tmp_path, getenv("HOME"));
		strcat(tmp_path, &(s[1]));
		s=tmp_path;
	}

	h=dlopen(s, RTLD_LAZY | RTLD_GLOBAL);
	if (h!=NULL) {
		void (*reg_savers)(void);
		reg_savers=dlsym(h, "reg_savers");
		error=dlerror();
		if (error) {
			fprintf(stderr, "%s\n", error);
			dlclose(h);
		} else {
			(*reg_savers)();
			if (m_list_size==0) {
				m_list_size=10;
				m_list=malloc(sizeof(char *)*m_list_size);
			}
			if (m_list_num==m_list_size) {
				m_list_size+=10;
				m_list=realloc(&m_list, sizeof(char *)*m_list_size);
			}
			m_list[m_list_num]=malloc(strlen(s)+1);
			strcpy(m_list[m_list_num], s);
			m_list_num++;
		}
	} else {
		error=dlerror();
		if (error)
			fprintf(stderr, "%s\n", error);
	}
}

void scan_modules(char *s)
{
	DIR *dir;
	struct dirent *entry;
	char name[255];
	
	dir=opendir(s);
	if (dir==NULL) {
		fprintf(stderr, "Sorry, unable to open %s !\n", s);
		return;
	}
	while ((entry=readdir(dir))!=NULL) {
                if ((strcmp(entry->d_name, ".")!=0) && (strcmp(entry->d_name, "..")!=0)) {
			sprintf(name, "%s/%s", s, entry->d_name);
			read_module(name); 
		}
        }
        closedir(dir);	
}

void free_modules()
{
	int i;
	for (i=0; i<m_list_num; i++) {
		dlclose(m_list[i]);
		free(m_list[i]);
	}
}

/*
 * screensaver logics:
 * (Needs a major cleanup and feature-enhancement.)
 */

int timeout=1;

ggi_visual_t visual;

ggi_color map[256];

int xsize,ysize;

int getic(void)
{ FILE *hand;
  int num,cnt;
  int c2=0;
  if ((hand=fopen("/proc/interrupts","r"))==NULL) 
    { perror("open /proc/interrupts"); 
      ggiPanic("Do you have /proc mounted ?\n"); }
  while(!feof(hand))
  { fscanf(hand,"%d:%d%*[^\n]",&num,&cnt);
    switch(num) /* Add other interrupts as you want it ... */
    { case 1: /* Keyboard */
      case 3: /* Mice on Com2/Com1 */
      case 4: c2+=cnt;break;
    }
  }
  fclose(hand);
  return(c2);
}

volatile int switchaway,switchto;

#if 0
void MyEvHand(struct ggi_event ev)
{ if (ev.evdata.vt.subtype==VT_SWITCH_AWAY) switchaway++;
  if (ev.evdata.vt.subtype==VT_SWITCH_TO) switchto++; }
#endif

/* These are the graphics effects displayed when active */

int SaverActive;

void do_saver(void)
{
	int oldvt;
	switchaway=0;
#if 0
	if ((oldvt=ggi_get_active_vt())<=0) return;
	ggi_vt_activate(0);
	ggiSetGCForeground(visual,0);ggiFillscreen(visual);
	rs_list[SaverActive].func(visual);
	while(ggiKbhit(visual)) ggiGetc(visual);
	if (!switchaway) ggi_vt_activate(oldvt);
#else
	rs_list[SaverActive].func(visual);
#endif
}

void blank_screen2(int interactive)
{ 
	int c;
	char hlpbuf[128];
	ggi_mode mode;
  
/*  ggi_register_evh(EV_ALL,MyEvHand);*/

	if ((visual=ggiOpen(NULL))==NULL) { 
		fprintf(stderr,"cannot open device.\n");
		exit(1);
	}

	ggiSetFocus(visual);

	if (ggiSetGraphMode(visual,GGI_AUTO,GGI_AUTO,GGI_AUTO,GGI_AUTO,GT_8BIT))
		ggiPanic("Cannot open one of the default modes.");

	ggiGetMode(visual,&mode);
	xsize=mode.visible.x;
	ysize=mode.visible.y;
      
	for(c=1;c<256;c++) {
		map[c].r=63*8*abs(c-128);
		map[c].b=63*7*(128-abs(c-128));
		map[c].g=63*100;
	}
	map[0].r=map[0].g=map[0].b=0;

	if (!interactive) { 
		ggiSetPaletteVec(visual,0,256,map);
		ggiFillscreen(visual);
		do_saver(); 
		return; 
	} else { 
		switchaway=0;
		while(!switchaway) { 
			ggiSetPaletteVec(visual,0,256,map);
			ggiFillscreen(visual);
			for(c=0;c<ysize;c++) {
				ggiSetGCForeground(visual,c/2+1);
				ggiDrawHLine(visual,0,c,xsize); 
			}
			ggiSetGCForeground(visual,128);
			ggiPuts(visual,10,10, "*** Screen - Saver ***");
			ggiPuts(visual,10,30, "Configuration Screen :");
			sprintf(hlpbuf,"n/p Type: %15s",rs_list[SaverActive].name);
			ggiPuts(visual,10,50, hlpbuf);
			sprintf(hlpbuf,"+/- Time: %4d minutes",timeout);
			ggiPuts(visual,10,70, hlpbuf);
			ggiPuts(visual,10,100,"Switch away to activate");
			while(!ggiKbhit(visual));
			switch(KTYP(U(c=ggiGetc(visual)))) {
				case KT_LATIN:
				case KT_LETTER: switch(KVAL(U(c))) {
					case '+': if (timeout<1440) timeout++;break;
					case '-': if (timeout>   1) timeout--;break;
					case 'p': if (SaverActive>0) SaverActive--;break;
					case 'n': if ((SaverActive+1)<rs_list_num) SaverActive++;break;
					case 't':
						ggiSetGCForeground(visual,0);
						ggiFillscreen(visual);
						do_saver();
						break;
					case '\x1b':
					case 'q': 
						ggiSetGCForeground(visual,0);
						ggiClose(visual);
      					      	ggiExit();
      					      	exit(0);
      					}
					break;
				default:  printf("Sym is %x.\n",c);break;
			}
		}
	}
	ggiClose(visual);
}

/*
 * MAIN
 */
int main(int argc,char *argv[])
{ 
	int i;
	int ic,ic2,cnt,x;
	rs_list_num = rs_list_size = m_list_num = m_list_size =0;

	ggiInit();


	/* 
	 * You have to fix this to the place, where you installed the 
	 * modules.
	 */
	scan_modules("/usr/src/saver-0.1/modules");

/*	Final version should do this: 
	scan_modules("/usr/lib/screensaver");
	scan_modules("/usr/local/lib/screensaver");
	scan_modules("~/.screensaver");
*/
	ic=cnt=0;
	switchto=1;

	if (argc>=2) {
		x=atoi(argv[1]);switchto=0;
		if (x>0&&x<1440 /* 1 Day ... */) timeout=x;
	}

	if (argc>=3) {
		x=atoi(argv[2]);
		if ((x>=0) && (x<rs_list_num))
			SaverActive=x;
	}
         
	while(1) { 
		if ((ic2=getic())==ic) {
			if (++cnt>=timeout) {
				blank_screen2(0);
				switchto=0;cnt=0;
			} 
		} else {
			cnt=0;
			ic=ic2;
		}
    
		if (!switchto) sleep(1/*60*/);
		if (switchto) {
			blank_screen2(1);
			switchto=0;
		}
	}

	free_modules();
	ggiClose(visual);
	ggiExit();
	return(0);
}
